const PART1: &str = include_str!("../data/day1a.txt");

pub fn part1() -> String {
    format!("{}", max_calories(sum_calories(input_data(PART1))).1)
}

pub fn part2() -> String {
    format!("{}", max_three_calories(sum_calories(input_data(PART1))))
}

fn input_data(data: &str) -> Vec<Vec<u32>> {
    data.lines()
        .collect::<Vec<&str>>()
        .split(|l| *l == "")
        .map(|v| {
            v.into_iter()
                .map(|element| element.parse::<u32>().unwrap())
                .collect::<Vec<u32>>()
        })
        .collect::<Vec<Vec<u32>>>()
}

fn sum_calories(food_lists: Vec<Vec<u32>>) -> Vec<(usize, u32)> {
    food_lists
        .into_iter()
        .map(|list| list.into_iter().fold(0, |cur, val| cur + val))
        .enumerate()
        .collect::<Vec<(usize, u32)>>()
}

fn elf_with_most_calories(batch_calories: Vec<(usize, u32)>) -> usize {
    let (idx, _) = max_calories(batch_calories);
    // Elves are 1-indexed
    return idx + 1;
}

fn max_calories(mut batch_calories: Vec<(usize, u32)>) -> (usize, u32) {
    batch_calories.sort_by(|(_, val1), (_, val2)| val1.cmp(val2));
    batch_calories.reverse();
    batch_calories[0]
}

fn max_three_calories(mut batch_calories: Vec<(usize, u32)>) -> u32 {
    batch_calories.sort_by(|(_, val1), (_, val2)| val1.cmp(val2));
    batch_calories.reverse();
    batch_calories
        .into_iter()
        .take(3)
        .fold(0, |cur, (_, val)| cur + val)
}

#[cfg(test)]
mod test {
    use super::*;

    const TEST_DATA1: &str = "1000
2000
3000

4000

5000
6000

7000
8000
9000

10000";
    #[test]
    fn find_most_elf() {
        let input = input_data(TEST_DATA1);
        let food_batches = sum_calories(input);
        assert_eq!(
            food_batches,
            vec![6000, 4000, 11000, 24000, 10000]
                .into_iter()
                .enumerate()
                .collect::<Vec<(usize, u32)>>()
        );

        assert_eq!(elf_with_most_calories(food_batches.clone()), 4);
        assert_eq!(max_calories(food_batches), (3, 24000));
    }

    #[test]
    fn batch_with_one_element() {
        let input = input_data("50057");
        let food_batches = sum_calories(input);
        assert_eq!(food_batches, vec![(0, 50057)]);
    }

    #[test]
    fn find_max_three() {
        assert_eq!(
            max_three_calories(sum_calories(input_data(TEST_DATA1))),
            45000
        );
    }
}
