use std::collections::HashSet;
use substring::Substring;

const INPUT: &str = include_str!("../data/day6.txt");

pub fn part1() -> String {
    format!("{:?}", start_of_packet(INPUT))
}

pub fn part2() -> String {
    format!("{:?}", start_of_message(INPUT))
}

fn start_of_packet(input: &str) -> Option<usize> {
    first_unique_window(input, 4)
}

fn start_of_message(input: &str) -> Option<usize> {
    first_unique_window(input, 14)
}

fn first_unique_window(input: &str, window_size: usize) -> Option<usize> {
    for window_idx in 0..(input.len() - window_size) {
        let window = input
            .substring(window_idx, window_idx + window_size)
            .chars()
            .collect::<HashSet<char>>();
        if window.len() == window_size {
            return Some(window_idx + window_size);
        }
    }
    None
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn it_finds_start_of_packet() {
        assert_eq!(start_of_packet("mjqjpqmgbljsphdztnvjfqwrcgsmlb"), Some(7));
        assert_eq!(start_of_packet("bvwbjplbgvbhsrlpgdmjqwftvncz"), Some(5));
        assert_eq!(start_of_packet("nppdvjthqldpwncqszvftbrmjlhg"), Some(6));
        assert_eq!(
            start_of_packet("nznrnfrfntjfmvfwmzdfjlvtqnbhcprsg"),
            Some(10)
        );
        assert_eq!(
            start_of_packet("zcfzfwzzqfrljwzlrfnpqdbhtmscgvjw"),
            Some(11)
        );
    }

    #[test]
    fn it_finds_start_of_message() {
        assert_eq!(start_of_message("mjqjpqmgbljsphdztnvjfqwrcgsmlb"), Some(19));
        assert_eq!(start_of_message("bvwbjplbgvbhsrlpgdmjqwftvncz"), Some(23));
        assert_eq!(start_of_message("nppdvjthqldpwncqszvftbrmjlhg"), Some(23));
        assert_eq!(
            start_of_message("nznrnfrfntjfmvfwmzdfjlvtqnbhcprsg"),
            Some(29)
        );
        assert_eq!(
            start_of_message("zcfzfwzzqfrljwzlrfnpqdbhtmscgvjw"),
            Some(26)
        );
    }
}
