const INPUT: &str = include_str!("../data/day1.txt");

pub fn day1a() -> String {
    format!("{}", get_calibration(INPUT, calibration_line))
}

pub fn day1b() -> String {
    format!("{}", get_calibration(INPUT, calibration_line_2))
}

pub(crate) fn get_calibration<F>(data: &str, parser: F) -> u32
where
    F: Fn(&str) -> u32,
{
    data.lines().map(parser).fold(0, |acc, val| acc + val)
}

pub(crate) fn calibration_line(data: &str) -> u32 {
    let chars = data
        .chars()
        .filter(|c| c.is_digit(10))
        .collect::<Vec<char>>();
    let number_as_str = format!("{}{}", chars[0], chars[chars.len() - 1]);
    number_as_str.parse::<u32>().unwrap()
}

pub(crate) fn calibration_line_2(data: &str) -> u32 {
    let data = interpret_text(data);
    calibration_line(&data)
}

const NAMES: [(&str, &str); 20] = [
    ("zero", "0"),
    ("one", "1"),
    ("two", "2"),
    ("three", "3"),
    ("four", "4"),
    ("five", "5"),
    ("six", "6"),
    ("seven", "7"),
    ("eight", "8"),
    ("nine", "9"),
    ("0", "0"),
    ("1", "1"),
    ("2", "2"),
    ("3", "3"),
    ("4", "4"),
    ("5", "5"),
    ("6", "6"),
    ("7", "7"),
    ("8", "8"),
    ("9", "9"),
];

fn interpret_text(data: &str) -> String {
    let mut working = data.to_owned();
    let mut digits = Vec::new();

    while let Some((idx, digit)) = find_digit(&working) {
        let _ = working.drain(..(idx + 1)).collect::<String>();
        digits.push(digit);
    }
    /*
    while let Some((idx, digit)) = find_digit(working.drain(..next_idx).collect()) {
        next_idx = idx;
        digits.push(digit);
    }
    */

    println!("interpretation {} -> {}", data, digits.join(""));
    digits.join("")
}

fn find_digit(data: &str) -> Option<(usize, String)> {
    let mut lowest_idx = 999;
    let mut first_value = None;

    for (name, value) in NAMES {
        if let Some(idx) = data.find(name) {
            if idx < lowest_idx {
                lowest_idx = idx;
                first_value = Some(value.to_owned());
            }
        }
    }

    first_value.map(|v| (lowest_idx, v))
}

/*
fn replace_first(data: String) -> (bool, String) {
    let names_hash = HashMap::from(NAMES);
    let mut lowest_idx = 999;
    let mut first_name = None;

    for (name, _) in NAMES {
        if let Some(idx) = data.find(name) {
            if idx < lowest_idx {
                lowest_idx = idx;
                first_name = Some(name);
            }
        }
    }

    if let Some(name) = first_name {
        return (true, data.replace(name, names_hash[name]));
    }
    return (false, data);
}
*/

#[cfg(test)]
mod test {
    use super::*;

    const TEST_DATA1: &str = "1abc2
pqr3stu8vwx
a1b2c3d4e5f
treb7uchet";

    const TEST_DATA2: &str = "two1nine
eightwothree
abcone2threexyz
xtwone3four
4nineeightseven2
zoneight234
7pqrstsixteen";

    #[test]
    fn test_1a() {
        let test_cases = vec![
            ("1abc2", 12, "1abc2"),
            ("pqr3stu8vwx", 38, "pqr3stu8vwx"),
            ("a1b2c3d4e5f", 15, "a1b2c3d4e5f"),
            ("treb7uchet", 77, "treb7uchet"),
        ];
        for (input, expected, message) in test_cases {
            assert_eq!(calibration_line(input), expected, "{}", message);
        }

        assert_eq!(get_calibration(TEST_DATA1, calibration_line), 142);
    }

    #[test]
    fn test_1b() {
        let test_cases = vec![
            ("two1nine", 29, "two1nine"),
            ("eightwothree", 83, "eightwothree"),
            ("abcone2threexyz", 13, "abcone2threexyz"),
            ("xtwone3four", 24, "xtwone3four"),
            ("4nineeightseven2", 42, "4nineeightseven2"),
            ("zoneight234", 14, "zoneight234"),
            ("7pqrstsixteen", 76, "7pqrstsixteen,"),
        ];
        for (input, expected, message) in test_cases {
            assert_eq!(calibration_line_2(input), expected, "{}", message);
        }

        assert_eq!(get_calibration(TEST_DATA2, calibration_line_2), 281);
    }

    #[test]
    fn test_interpret_text() {
        let test_cases = vec![
            ("two1nine", "219", "two1nine"),
            ("eightwothree", "823", "eightwothree"),
            ("abcone2threexyz", "123", "abcone2threexyz"),
            ("xtwone3four", "2134", "xtwone3four"),
            ("4nineeightseven2", "49872", "4nineeightseven2"),
            ("zoneight234", "18234", "zoneight234"),
            ("7pqrstsixteen", "76", "7pqrstsixteen,"),
        ];

        for (input, expected, message) in test_cases {
            assert_eq!(interpret_text(input), expected, "{}", message);
        }
    }
}
